<?php
/************************************************************************
 * Mailer v0.2.1-FINAL                                Start: 08/29/2004 *
 * ===================                          Last change: 08/29/2004 *
 *                                                                      *
 * -------------------------------------------------------------------- *
 * File              : lib-mysql3.php                                   *
 * -------------------------------------------------------------------- *
 * Short description : Database layer for MySQL 3/4/5 server            *
 * -------------------------------------------------------------------- *
 * Kurzbeschreibung  : Datenbankschicht fuer MySQL 3/4/5 Server         *
 * -------------------------------------------------------------------- *
 * $Revision::                                                        $ *
 * $Date::                                                            $ *
 * $Tag:: 0.2.1-FINAL                                                 $ *
 * $Author::                                                          $ *
 * -------------------------------------------------------------------- *
 * Copyright (c) 2003 - 2009 by Roland Haeder                           *
 * Copyright (c) 2009, 2010 by Mailer Developer Team                    *
 * For more information visit: http://www.mxchange.org                  *
 *                                                                      *
 * This program is free software; you can redistribute it and/or modify *
 * it under the terms of the GNU General Public License as published by *
 * the Free Software Foundation; either version 2 of the License, or    *
 * (at your option) any later version.                                  *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,               *
 * MA  02110-1301  USA                                                  *
 ************************************************************************/

// Some security stuff...
if (!defined('__SECURITY')) {
	die();
} // END - if

// SQL queries
function SQL_QUERY ($sqlString, $F, $L, $enableCodes = true) {
	// Do we have cache?
	if (!isset($GLOBALS[__FUNCTION__][$sqlString])) {
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Called: ' . $sqlString);

		// Trim SQL string
		$sqlStringModified = trim($sqlString);

		// Empty query string or link is not up?
		if (empty($sqlStringModified)) {
			// Empty SQL string!
			debug_report_bug(__FUNCTION__, __LINE__, sprintf("SQL string is empty. Please fix this. file=%s, line=%s",
				basename($F),
				$L
			));
		} elseif (!SQL_IS_LINK_UP()) {
			// We should not quietly ignore this
			debug_report_bug(__FUNCTION__, __LINE__, sprintf("Cannot query database: sqlString=%s,file=%s,line=%s",
				$sqlStringModified,
				basename($F),
				$L
			));
		}

		// Remove \t, \n and \r from queries they may confuse some MySQL versions
		$sqlStringModified = str_replace("\t", ' ', str_replace("\n", ' ', str_replace("\r", ' ', $sqlStringModified)));

		// Compile config entries out
		$sqlStringModified = SQL_PREPARE_SQL_STRING($sqlStringModified, $enableCodes);

		// Cache it and remember as last SQL query
		$GLOBALS[__FUNCTION__][$sqlString] = $sqlStringModified;
		$GLOBALS['last_sql'] = $sqlStringModified;
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Stored cache: ' . $sqlStringModified);
	}  else {
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Cache used: ' . $sqlString);

		// Use cache (to save a lot function calls
		$GLOBALS['last_sql'] = $GLOBALS[__FUNCTION__][$sqlString];

		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Cache is: ' . $sqlString);
	}

	// Starting time
	$querytimeBefore = microtime(true);

	// Run SQL command
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'F=' . basename($F) . ',L=' . $L . ',sql=' . $GLOBALS['last_sql']);
	$result = mysql_query($GLOBALS['last_sql'], SQL_GET_LINK())
		or debug_report_bug($F, $L, 'file='. basename($F) . ',line=' . $L . ':mysql_error()=' . mysql_error() . "\n".
'Query string:' . $GLOBALS['last_sql']);
	//* DEBUG: */ logDebugMessage($F, $L, 'sql=' . $GLOBALS['last_sql'] . ',affected=' . SQL_AFFECTEDROWS() . ',numRows='.(is_resource($result) ? SQL_NUMROWS($result) : gettype($result)));

	// Calculate query time
	$queryTime = microtime(true) - $querytimeBefore;

	// Add this query to array including timing
	addSqlToDebug($result, $GLOBALS['last_sql'], $queryTime, $F, $L);

	// Save last successfull query
	setConfigEntry('db_last_query', $GLOBALS['last_sql']);

	// Count all query times
	incrementConfigEntry('sql_time', $queryTime);

	// Count this query
	incrementConfigEntry('sql_count');

	// Debug output
	if ((!isCssOutputMode()) && (isDebugModeEnabled()) && (isSqlDebuggingEnabled())) {
		// Is this the first call?
		if (!isset($GLOBALS['sql_first_entry'])) {
			// Write first entry
			appendLineToFile(getCachePath() . 'mysql.log', 'Module=' . getModule());
			$GLOBALS['sql_first_entry'] = true;
		} // END - if

		// Append debug line
		appendLineToFile(getCachePath() . 'mysql.log', $F . '(LINE=' . $L . '|NUM=' . SQL_NUMROWS($result) . '|AFFECTED=' . SQL_AFFECTEDROWS() . '|QUERYTIME:' . $queryTime . '): ' . str_replace("\r", '', str_replace("\n", ' ', $GLOBALS['last_sql'])));
	} // END - if

	// Count DB hits
	if (!isStatsEntrySet('db_hits')) {
		// Count in dummy variable
		setStatsEntry('db_hits', 1);
	} else {
		// Count to config array
		incrementStatsEntry('db_hits');
	}

	// Return the result
	return $result;
}

// SQL num rows
function SQL_NUMROWS ($resource) {
	// Valid link resource?
	if (!SQL_IS_LINK_UP()) return false;

	// Link is not up, no rows by default
	$lines = false;

	// Is the result a valid resource?
	if (isset($GLOBALS['sql_numrows'][$resource])) {
		// Use cache
		$lines = $GLOBALS['sql_numrows'][intval($resource)];
	} elseif (is_resource($resource)) {
		// Get the count of rows from database
		$lines = mysql_num_rows($resource);

		// Remember it in cache
		$GLOBALS['sql_numrows'][intval($resource)] = $lines;
	} else {
		// No resource given, please fix this
		debug_report_bug(__FUNCTION__, __LINE__, 'No resource given! result[]=' . gettype($resource) . ',last_sql=' .  $GLOBALS['last_sql']);
	}

	// Return lines
	return $lines;
}

// SQL affected rows
function SQL_AFFECTEDROWS() {
	// Valid link resource?
	if (!SQL_IS_LINK_UP()) return false;

	// Get affected rows
	$lines = mysql_affected_rows(SQL_GET_LINK());

	// Return it
	return $lines;
}

// SQL fetch row
function SQL_FETCHROW ($resource) {
	// Is a result resource set?
	if ((!is_resource($resource)) || (!SQL_IS_LINK_UP())) return false;

	// Fetch the data and return it
	return mysql_fetch_row($resource);
}

// SQL fetch array
function SQL_FETCHARRAY ($res) {
	// Is a result resource set?
	if ((!is_resource($res)) || (!SQL_IS_LINK_UP())) return false;

	// Load row from database
	$row = mysql_fetch_assoc($res);

	// Return only arrays here
	if (is_array($row)) {
		// Return row
		return $row;
	} else {
		// Return a false, else some loops would go endless...
		return false;
	}
}

// SQL result
function SQL_RESULT ($resource, $row, $field = '0') {
	// Is $resource valid?
	if ((!is_resource($resource)) || (!SQL_IS_LINK_UP())) return false;

	// Run the result command
	$result = mysql_result($resource, $row, $field);

	// ... and return the result
	return $result;
}

// SQL connect
function SQL_CONNECT ($host, $login, $password, $F, $L) {
	// Try to connect
	$linkResource = mysql_connect($host, $login, $password) or addFatalMessage(__FUNCTION__, __LINE__, $F . ' (' . $L . '):' . mysql_error());

	// Set the link resource
	SQL_SET_LINK($linkResource);

	// Destroy cache
	unset($GLOBALS['is_sql_link_up']);
}

// SQL select database
function SQL_SELECT_DB ($dbName, $F, $L) {
	// Is there still a valid link? If not, skip it.
	if (!SQL_IS_LINK_UP()) return false;

	// Return the result
	return mysql_select_db($dbName, SQL_GET_LINK()) or addFatalMessage(__FUNCTION__, __LINE__, $F . ' (' . $L . '):' . mysql_error());
}

// SQL close link
function SQL_CLOSE ($F, $L) {
	if (!SQL_IS_LINK_UP()) {
		// Skip double close
		return false;
	} // END - if

	// Close database link and forget the link
	$close = mysql_close(SQL_GET_LINK())
		or addFatalMessage(__FUNCTION__, __LINE__, $F . ' (' . $L . '):'.mysql_error());

	// Close link
	SQL_SET_LINK(null);

	// Destroy cache
	unset($GLOBALS['is_sql_link_up']);

	// Return the result
	return $close;
}

// SQL free result
function SQL_FREERESULT ($resource) {
	if ((!is_resource($resource)) || (!SQL_IS_LINK_UP())) {
		// Abort here
		return false;
	} // END - if

	// Free result
	$res = mysql_free_result($resource);

	// And return that result of freeing it...
	return $res;
}

// SQL string escaping
function SQL_QUERY_ESC ($sqlString, $data, $F, $L, $run = true, $strip = true, $secure = true) {
	// Link is there?
	if ((!SQL_IS_LINK_UP()) || (!is_array($data))) return false;

	// Escape all data
	$dataSecured['__sql_string'] = $sqlString;
	foreach ($data as $key => $value) {
		$dataSecured[$key] = SQL_ESCAPE($value, $secure, $strip);
	} // END - foreach

	// Generate query
	$query = call_user_func_array('sprintf', $dataSecured);

	if ($run === true) {
		// Run SQL query (default)
		return SQL_QUERY($query, $F, $L);
	} else {
		// Return secured string
		return $query;
	}
}

// Get id from last INSERT command
function SQL_INSERTID () {
	if (!SQL_IS_LINK_UP()) return false;
	return mysql_insert_id();
}

// Escape a string for the database
function SQL_ESCAPE ($str, $secureString = true, $strip = true) {
	// Do we have cache?
	if (!isset($GLOBALS['sql_escapes'][''.$str.''])) {
		// Secure string first? (which is the default behaviour!)
		if ($secureString === true) {
			// Then do it here
			$str = secureString($str, $strip);
		} // END - if

		if (!SQL_IS_LINK_UP()) {
			// Fall-back to escapeQuotes() when there is no link
			$ret = escapeQuotes($str);
		} elseif (function_exists('mysql_real_escape_string')) {
			// The new and improved version
			$ret = mysql_real_escape_string($str, SQL_GET_LINK());
		} elseif (function_exists('mysql_escape_string')) {
			// The obsolete function
			$ret = mysql_escape_string($str, SQL_GET_LINK());
		} else {
			// If nothing else works, fall back to escapeQuotes() again
			$ret = escapeQuotes($str);
		}

		// Cache result
		$GLOBALS['sql_escapes'][''.$str.''] = $ret;
	} // END - if

	// Return it
	return $GLOBALS['sql_escapes'][''.$str.''];
}

// SELECT query string from table, columns and so on... ;-)
function SQL_RESULT_FROM_ARRAY ($table, $columns, $idRow, $id, $F, $L) {
	// Is columns an array?
	if (!is_array($columns)) {
		// No array
		debug_report_bug(__FUNCTION__, __LINE__, sprintf("columns is not an array. %s != array, file=%s, line=%s",
			gettype($columns),
			basename($F),
			$L
		));

		// Abort here with 'false'
		return false;
	} // END  - if

	// Prepare the SQL statement
	$sql = "SELECT `".implode("`,`", $columns)."` FROM `{?_MYSQL_PREFIX?}_%s` WHERE `%s`='%s' LIMIT 1";

	// Return the result
	return SQL_QUERY_ESC($sql,
		array(
			$table,
			$idRow,
			bigintval($id),
		), $F, $L
	);
}

// ALTER TABLE wrapper function
function SQL_ALTER_TABLE ($sql, $F, $L, $enableCodes = true) {
	// Abort if link is down
	if (!SQL_IS_LINK_UP()) return false;

	// This is the default result...
	$result = false;

	// Determine index/fulltext/unique word
	$noIndex = (
	(
		strpos($sql, 'INDEX') === false
	) && (
		strpos($sql, 'KEY') === false
	) && (
		strpos($sql, 'FULLTEXT') === false
	) && (
		strpos($sql, 'UNIQUE') === false
	)
	);

	// Extract table name
	$tableArray = explode(' ', $sql);
	$tableName = str_replace('`', '', $tableArray[2]);

	// Debug log
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'sql=' . $sql . ',tableName=' . $tableName);

	// Shall we add/drop?
	if (((strpos($sql, 'ADD') !== false) || (strpos($sql, 'DROP') !== false) || (strpos($sql, 'CHANGE') !== false)) && ($noIndex === true)) {
		// Try two columns, one should fix
		foreach (array(4,5) as $idx) {
			// If an entry is not set, abort here
			if (!isset($tableArray[$idx])) {
				// Debug log this
				logDebugMessage(__FUNCTION__, __LINE__, 'columnName=' . $columnName . ',idx=' . $idx . ',sql=' . $sql . ' is missing!');
				break;
			} // END - if

			// And column name as well
			$columnName = str_replace('`', '', $tableArray[$idx]);

			// Get column information
			$result = SQL_QUERY_ESC("SHOW COLUMNS FROM `%s` LIKE '%s'",
				array($tableName, $columnName), __FUNCTION__, __LINE__);

			// Debug log
			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'columnName=' . $columnName . ',idx=' . $idx . ',sql=' . $sql . ',hasZeroNums=' . intval(SQL_HASZERONUMS($result)));

			// Do we have no entry on ADD or an entry on DROP/CHANGE?
			// 123               4       43    3      4           4          32    23           4       4     3    3      4            4          32    23                4       43    3      4              4          3    3         321
			if (((SQL_HASZERONUMS($result)) && (strpos($sql, 'ADD') !== false)) || ((!SQL_HASZERONUMS($result)) && (strpos($sql, 'DROP') !== false)) || ((!SQL_HASZERONUMS($result)) && (strpos($sql, 'CHANGE') !== false) && ($idx == 4))) {
				// Do the query
				//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Executing: ' . $sql);
				$result = SQL_QUERY($sql, $F, $L, false);

				// Skip further attempt(s)
				break;
			//       1234                5       54    4      5           5          43    34                5       54    4      5            5          43    3      4              4          32    2                    21
			} elseif ((((!SQL_HASZERONUMS($result)) && (strpos($sql, 'ADD') !== false)) || ((!SQL_HASZERONUMS($result)) && (strpos($sql, 'DROP') !== false)) || (strpos($sql, 'CHANGE') !== false)) && ($columnName != 'KEY')) {
				// Abort here because it is alreay there
				//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Skipped: sql=' . $sql . ',columnName=' . $columnName . ',idx=' . $idx);
				break;
			} elseif ((SQL_HASZERONUMS($result)) && (strpos($sql, 'DROP') !== false)) {
				// Abort here because we tried to drop a column which is not there (never created maybe)
				//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'No drop: ' . $sql);
				break;
			} elseif ($columnName != 'KEY') {
				// Something didn't fit, we better log it
				logDebugMessage(__FUNCTION__, __LINE__, 'Possible problem: ' . $sql . ',hasZeroNums=' . intval(SQL_HASZERONUMS($result)) . '');
			}
		} // END - foreach
	} elseif ((getConfig('_TABLE_TYPE') == 'InnoDB') && (strpos($sql, 'FULLTEXT') !== false)) {
		// Skip this query silently
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, sprintf("Skipped FULLTEXT: sql=%s,tableName=%s,hasZeroNums=%d,file=%s,line=%s", $sql, $tableName, intval((is_bool($result)) ? 0 : SQL_HASZERONUMS($result)), $F, $L));
	} elseif ($noIndex === false) {
		// And column name as well
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'tableArray=<pre>' . print_r($tableArray, true) . '</pre>');
		$keyName = str_replace('`', '', $tableArray[5]);

		// Is this "UNIQUE" or so? FULLTEXT has been handled the elseif() block above
		if (in_array(strtoupper($keyName), array('INDEX', 'UNIQUE', 'KEY', 'FULLTEXT'))) {
			// Init loop
			$begin = 1; $keyName = ',';
			while (strpos($keyName, ',') !== false) {
				// Use last
				$keyName = str_replace('`', '', $tableArray[count($tableArray) - $begin]);
				//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, $keyName . '----------------' . $begin);

				// Remove brackes
				$keyName = str_replace('(', '', str_replace(')', '', $keyName));
				//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, $keyName . '----------------' . $begin);

				// Continue
				$begin++;
			} // END while
		} // END - if

		// Show indexes
		$result = SQL_QUERY_ESC("SHOW INDEX FROM `%s`", array($tableName), __FUNCTION__, __LINE__);

		// Non-skipping is default for ADD
		$skip = false;

		// But should we DROP?
		if ($tableArray[3] == 'DROP') {
			// Then skip if nothing found!
			$skip = true;
			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Going to drop key ' . $keyName);
		} // END - if

		// Walk through all
		while ($content = SQL_FETCHARRAY($result)) {
			// Is it found?
			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'columnName='.$keyName.',content=<pre>' . print_r($content, true) . '</pre>');
			if (($content['Key_name'] == $keyName) && ($tableArray[3] == 'ADD')) {
				// Skip this query!
				//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, sprintf("ADD: Skiped: %s", $sql));
				$skip = true;
				break;
			} elseif (($content['Key_name'] == $keyName) && ($tableArray[3] == 'DROP')) {
				// Don't skip this!
				//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, sprintf("DROP: Not skiped: %s", $sql));
				$skip = false;
				break;
			}
		} // END - while

		// Free result
		SQL_FREERESULT($result);

		// Shall we run it?
		if ($skip === false) {
			// Send it to the SQL_QUERY() function
			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, $sql);
			$result = SQL_QUERY($sql, $F, $L, $enableCodes);
		} else {
			// Not executed
			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Not executed: ' . $sql);
		}
	} else {
		// Other ALTER TABLE query
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, $sql);
		$result = SQL_QUERY($sql, $F, $L, $enableCodes);
	}

	// Return result
	return $result;
}

// Getter for SQL link
function SQL_GET_LINK () {
	// Init link
	$link = null;

	// Is it in the globals?
	if (isset($GLOBALS['sql_link'])) {
		// Then take it
		$link = $GLOBALS['sql_link'];
	} // END - if

	// Return it
	return $link;
}

// Setter for link
function SQL_SET_LINK ($link) {
	// Is this a resource or null?
	if ((ifFatalErrorsDetected()) && (isInstallationPhase())) {
		// This may happen in installation phase
		return;
	} elseif ((!is_resource($link)) && (!is_null($link))) {
		// This should never happen!
		debug_report_bug(__FUNCTION__, __LINE__, sprintf("link is not resource or null. Type: %s", gettype($link)));
	} // END - if

	// Set it
	$GLOBALS['sql_link'] = $link;
}

// Checks if the link is up
function SQL_IS_LINK_UP () {
	// Default is not up
	$linkUp = false;

	// Do we have cached this?
	if (isset($GLOBALS['is_sql_link_up'])) {
		// Then use this
		$linkUp = $GLOBALS['is_sql_link_up'];
	} else {
		// Get it
		$linkUp = is_resource(SQL_GET_LINK());

		// And cache it
		$GLOBALS['is_sql_link_up'] = $linkUp;
	}

	// Return the result
	return $linkUp;
}

// Wrapper function to make code more readable
function SQL_HASZERONUMS ($result) {
	// Just pass it through
	return (SQL_NUMROWS($result) === 0);
}

// Wrapper function to make code more readable
function SQL_HASZEROAFFECTED () {
	// Just pass it through
	return (SQL_AFFECTEDROWS() === 0);
}

// Private function to prepare the SQL query string
function SQL_PREPARE_SQL_STRING ($sqlString, $enableCodes = true) {
	// Is it already cached?
	if (!isset($GLOBALS['sql_strings'][$sqlString])) {
		// Compile config+expression code
		$sqlString2 = FILTER_COMPILE_EXPRESSION_CODE(FILTER_COMPILE_CONFIG($sqlString));

		// Do final compilation
		$GLOBALS['sql_strings'][$sqlString] = doFinalCompilation($sqlString2, false, $enableCodes);
		//die($sqlString.'<br />'.$sqlString2.'<br />'.$GLOBALS['sql_strings'][$sqlString]);
	} // END - if

	// Return it
	return $GLOBALS['sql_strings'][$sqlString];
}

// [EOF]
?>
